from dldstraining.core.exceptions import NotReferencable, ImplementationError

from typing import Any, List, Optional, Set, Union
from inspect import stack, signature


class Ref:
    def __init__(self, reference: Any, attribute: str):
        self.__reference: Any = reference
        self.__attribute: str = attribute
        if not hasattr(self.__reference, self.__attribute):
            raise ValueError(f'Reference {self.__reference} has no attribute {self.__attribute}')

    def __call__(self):
        return getattr(self.__reference, self.__attribute)


def ref(reference: Any, object: Any):
    for sub_var in filter(lambda va: getattr(reference, va) is object, vars(reference)):
        return Ref(reference, sub_var)
    raise ValueError(f'Invalid reference: {object} cannot be referenced from {reference}')


def to_reference(variable, stack_id: int = 1) -> Union[Ref, Any]:
    if callable(variable) or isinstance(variable, Ref):
        return variable

    def __find(og_variable, key, var, i: int = 2) -> Optional[Ref]:
        if hasattr(var, '__dict__') and not key.startswith('__') and i > 0:
            for sub_var in filter(lambda va: getattr(var, va) is og_variable, vars(var)):
                lr = Ref(var, sub_var)
                return lr

            for a in vars(var):
                rci = __find(og_variable, a, getattr(var, a), i - 1)
                if rci:
                    return rci

        return None

    for fa in filter(lambda a: not a[0].startswith('__') and hasattr(a[1], '__dict__'),
                     stack()[stack_id][0].f_locals.items()):

        rc = __find(variable, *fa)
        if rc:
            return rc

    raise NotReferencable(variable)


def as_reference(variable_names: Optional[Set[str]] = None):
    """
    Names of the variables that are to be passed as references

    :param variable_names:
    :return:
    """

    def reference_names(wrapped_function):

        function_args: List[str] = list(signature(wrapped_function).parameters.keys())

        # Threshold: check that the `variable_names` that are to be altered are valid arguments to the
        # wrapped function.
        if __debug__:
            missing_arguments: List[str] = [ma for ma in variable_names if ma not in function_args]
            assert not any(missing_arguments), (f'Invalid decorator argument{"s" * min(len(missing_arguments), 1)}: '
                                                f'{", ".join(missing_arguments)} to {wrapped_function.__name__}')

        def adapted(*args, **kwargs):
            full_args = dict(**kwargs, **{function_args[i]: args[i] for i in range(len(args))})
            for ra in variable_names if variable_names is not None else full_args.keys():
                if ra in full_args.keys():
                    if full_args[ra] is not None:
                        if type(full_args[ra]) == dict:
                            full_args[ra] = {k: to_reference(v, 3) for k, v in full_args[ra].items()}
                        elif type(full_args[ra]) == list:
                            full_args[ra] = [to_reference(a, 3) for a in full_args[ra]]
                        else:
                            full_args[ra] = to_reference(full_args[ra], 2)

            return wrapped_function(**full_args)

        return adapted

    return reference_names
